/*
    TreeP Run Time Support
    Copyright (C) 2008-2012 Frank Sinapsi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../trp/trp.h"
#include "./trpcv.h"
#include "../trppix/trppix_internal.h"
#include "sift/imgfeatures.h"
#include "sift/minpq.h"
#include "sift/utils.h"
#include "sift/kdtree.h"
#include "sift/sift.h"
#include "sift/xform.h"
#include <cv.h>
#include <highgui.h>

/* the maximum number of keypoint NN candidates to check during BBF search */
#define KDTREE_BBF_MAX_NN_CHKS 200

/* threshold on squared ratio of distances between NN and 2nd NN */
/* #define NN_SQ_DIST_RATIO_THR 0.49 */
#define NN_SQ_DIST_RATIO_THR 0.3

typedef struct {
    uns8b tipo;
    uns8b sottotipo;
    uns32b n;
    struct feature *feat;
} trp_cv_t;

uns32b trp_size_internal( trp_obj_t *obj );
void trp_encode_internal( trp_obj_t *obj, uns8b **buf );

static uns8b trp_cv_print( trp_print_t *p, trp_cv_t *obj );
static uns8b trp_cv_close( trp_cv_t *obj );
static uns8b trp_cv_close_basic( uns8b flags, trp_cv_t *obj );
static void trp_cv_finalize( void *obj, void *data );
static uns32b trp_cv_size( trp_cv_t *obj );
static void trp_cv_encode( trp_cv_t *obj, uns8b **buf );
static trp_obj_t *trp_cv_decode( uns8b **buf );
static trp_obj_t *trp_cv_equal( trp_cv_t *obj1, trp_cv_t *obj2 );
static trp_obj_t *trp_cv_length( trp_cv_t *obj );
static IplImage *trp_cv_pix2IplImage( trp_obj_t *pix );
static void trp_cv_copy_im2pix( IplImage *im, trp_obj_t *pix );
static uns8b trp_pix_load_cv( uns8b *cpath, uns32b *w, uns32b *h, uns8b **data );
static trp_obj_t *trp_cv_sift_features_basic( uns8b flags, trp_obj_t *pix );

uns8b trp_cv_init()
{
    extern uns8bfun_t _trp_pix_load_cv;
    extern uns8bfun_t _trp_print_fun[];
    extern uns8bfun_t _trp_close_fun[];
    extern uns32bfun_t _trp_size_fun[];
    extern voidfun_t _trp_encode_fun[];
    extern objfun_t _trp_decode_fun[];
    extern objfun_t _trp_equal_fun[];
    extern objfun_t _trp_length_fun[];

    _trp_pix_load_cv = trp_pix_load_cv;
    _trp_print_fun[ TRP_OPENCV ] = trp_cv_print;
    _trp_close_fun[ TRP_OPENCV ] = trp_cv_close;
    _trp_size_fun[ TRP_OPENCV ] = trp_cv_size;
    _trp_encode_fun[ TRP_OPENCV ] = trp_cv_encode;
    _trp_decode_fun[ TRP_OPENCV ] = trp_cv_decode;
    _trp_equal_fun[ TRP_OPENCV ] = trp_cv_equal;
    _trp_length_fun[ TRP_OPENCV ] = trp_cv_length;
    return 0;
}

static uns8b trp_cv_print( trp_print_t *p, trp_cv_t *obj )
{
    if ( trp_print_char_star( p, "#opencv" ) )
        return 1;
    if ( obj->feat == NULL )
        if ( trp_print_char_star( p, " (closed)" ) )
            return 1;
    return trp_print_char( p, '#' );
}

static uns8b trp_cv_close( trp_cv_t *obj )
{
    return trp_cv_close_basic( 1, obj );
}

static uns8b trp_cv_close_basic( uns8b flags, trp_cv_t *obj )
{
    if ( obj->feat ) {
        if ( flags & 1 )
            trp_gc_remove_finalizer( (trp_obj_t *)obj );
        free( obj->feat );
        obj->feat = NULL;
    }
    return 0;
}

static void trp_cv_finalize( void *obj, void *data )
{
    trp_cv_close_basic( 0, (trp_cv_t *)obj );
}

static uns32b trp_cv_size( trp_cv_t *obj )
{
    uns32b sz;

    /*
     FIXME
     */
    if ( obj->feat )
        sz = 1 + 1 + 4 + obj->n * sizeof( struct feature );
    else
        sz = trp_size_internal( UNDEF );
    return sz;
}

static void trp_cv_encode( trp_cv_t *obj, uns8b **buf )
{
    /*
     FIXME
     */
}

static trp_obj_t *trp_cv_decode( uns8b **buf )
{
    /*
     FIXME
     */
    return UNDEF;
}

static trp_obj_t *trp_cv_equal( trp_cv_t *obj1, trp_cv_t *obj2 )
{
    /*
     FIXME
     */
    return TRP_FALSE;
}

static trp_obj_t *trp_cv_length( trp_cv_t *obj )
{
    return obj->feat ? trp_sig64( obj->n ) : UNDEF;
}

trp_obj_t *trp_cv_version()
{
    uns8b buf[ 12 ];

    sprintf( buf, "%d.%d.%d",
             CV_MAJOR_VERSION,
             CV_MINOR_VERSION,
             CV_SUBMINOR_VERSION );
    return trp_cord( buf );
}

static IplImage *trp_cv_pix2IplImage( trp_obj_t *pix )
{
    IplImage *im;
    trp_pix_color_t *c;
    uns8b *p;
    uns32b w, h, i, j;

    if ( pix->tipo != TRP_PIX )
        return NULL;
    if ( ( c = ((trp_pix_t *)pix)->map.c ) == NULL )
        return NULL;
    w = ((trp_pix_t *)pix)->w;
    h = ((trp_pix_t *)pix)->h;
    if ( ( im = cvCreateImage( cvSize( w, h ), IPL_DEPTH_8U, 3 ) ) == NULL )
        return NULL;
    for ( i = 0 ; i < h ; i++ )
        for ( p = im->imageData + i * im->widthStep, j = 0 ; j < w ; j++, c++ ) {
            *p++ = c->blue;
            *p++ = c->green;
            *p++ = c->red;
        }
    return im;
}

static void trp_cv_copy_im2pix( IplImage *im, trp_obj_t *pix )
{
    trp_pix_color_t *c = ((trp_pix_t *)pix)->map.c;
    uns8b *p;
    uns32b w = ((trp_pix_t *)pix)->w, h = ((trp_pix_t *)pix)->h, i, j;

    switch ( im->nChannels ) {
    case 1:
        for ( i = 0 ; i < h ; i++ )
            for ( p = im->imageData + i * im->widthStep, j = 0 ; j < w ; j++, c++ ) {
                c->red = c->green = c->blue = *p++;
            }
        break;
    case 3:
        for ( i = 0 ; i < h ; i++ )
            for ( p = im->imageData + i * im->widthStep, j = 0 ; j < w ; j++, c++ ) {
                c->blue = *p++;
                c->green = *p++;
                c->red = *p++;
            }
        break;
    }
}

static uns8b trp_pix_load_cv( uns8b *cpath, uns32b *w, uns32b *h, uns8b **data )
{
    IplImage *im = cvLoadImage( cpath, 1 );
    trp_pix_color_t *c;
    uns8b *p;
    uns32b i, j;

    if ( im == NULL )
        return 1;
    /*
    printf( "#supp: %d x %d, depth=%d, ch=%d, alpha=%d\n",
            im->width, im->height, im->depth, im->nChannels, im->alphaChannel );
     */
    if ( ( im->depth != 8 ) || ( im->nChannels != 3 ) ) {
        cvReleaseImage( &im );
        return 1;
    }
    *w = im->width;
    *h = im->height;
    if ( ( *data = malloc( ( *w * *h ) << 2 ) ) == NULL ) {
        cvReleaseImage( &im );
        return 1;
    }
    c = (trp_pix_color_t *)( *data );
    for ( i = 0 ; i < *h ; i++ )
        for ( p = im->imageData + i * im->widthStep, j = 0 ; j < *w ; j++, c++ ) {
            c->alpha = 0xff;
            c->blue = *p++;
            c->green = *p++;
            c->red = *p++;
        }
    cvReleaseImage( &im );
    return 0;
}

trp_obj_t *trp_cv_pix_load( trp_obj_t *path )
{
    uns8b *cpath = trp_csprint( path ), *data;
    uns32b w, h;

    if ( trp_pix_load_cv( cpath, &w, &h, &data ) ) {
        trp_csprint_free( cpath );
        return UNDEF;
    }
    trp_csprint_free( cpath );
    return trp_pix_create_image_from_data( 0, w, h, data );
}

uns8b trp_cv_pix_gray( trp_obj_t *pix )
{
    IplImage *im = trp_cv_pix2IplImage( pix ), *gim;

    if ( im == NULL )
        return 1;
    if ( gim = cvCreateImage( cvGetSize( im ), IPL_DEPTH_8U, 1 ) )
        cvCvtColor( im, gim, CV_BGR2GRAY );
    cvReleaseImage( &im );
    if ( gim == NULL )
        return 1;
    trp_cv_copy_im2pix( gim, pix );
    cvReleaseImage( &gim );
    return 0;
}

static trp_obj_t *trp_cv_sift_features_basic( uns8b flags, trp_obj_t *pix )
{
    IplImage *im = trp_cv_pix2IplImage( pix );
    struct feature *feat;
    uns32b n;

    if ( im == NULL )
        return UNDEF;
    n = sift_features( im, &feat );
    if ( ( flags & 1 ) && n ) {
        draw_features( im, feat, n );
        trp_cv_copy_im2pix( im, pix );
    }
    cvReleaseImage( &im );
    if ( n == 0 )
        return UNDEF;
    pix = trp_gc_malloc_atomic_finalize( sizeof( trp_cv_t ), trp_cv_finalize );
    pix->tipo = TRP_OPENCV;
    ((trp_cv_t *)pix)->sottotipo = 0;
    ((trp_cv_t *)pix)->n = n;
    ((trp_cv_t *)pix)->feat = feat;
    return pix;
}

trp_obj_t *trp_cv_sift_features( trp_obj_t *pix )
{
    return trp_cv_sift_features_basic( 0, pix );
}

trp_obj_t *trp_cv_sift_features_draw( trp_obj_t *pix )
{
    return trp_cv_sift_features_basic( 1, pix );
}

trp_obj_t *trp_cv_sift_match( trp_obj_t *o1, trp_obj_t *o2, trp_obj_t *thr )
{
    trp_obj_t *res;
    struct feature *feat1, *feat2, *feat, **nbrs;
    struct kd_node *kd_root;
    double dthr;
    uns32b n1, n2, i, k;

    if ( thr == NULL )
        dthr = NN_SQ_DIST_RATIO_THR;
    else if ( trp_cast_double_range( thr, &dthr, 0.0, 1.0 ) )
        return UNDEF;
    if ( ( o1->tipo != TRP_OPENCV ) || ( o2->tipo != TRP_OPENCV ) )
        return UNDEF;
    if ( ( ( feat1 = ((trp_cv_t *)o1)->feat ) == NULL ) ||
         ( ( feat2 = ((trp_cv_t *)o2)->feat ) == NULL ) )
        return UNDEF;
    res = trp_queue();
    n1 = ((trp_cv_t *)o1)->n;
    n2 = ((trp_cv_t *)o2)->n;
    if ( ( n1 < 2 ) || ( n2 < 2 ) )
        return res;
    kd_root = kdtree_build( feat2, n2 );
    for ( i = 0 ; i < n1 ; i++ ) {
        feat = feat1 + i;
        k = kdtree_bbf_knn( kd_root, feat, 2, &nbrs, KDTREE_BBF_MAX_NN_CHKS );
        if( k == 2 ) {
            if ( descr_dist_sq( feat, nbrs[0] ) < descr_dist_sq( feat, nbrs[1] ) * dthr ) {
                trp_queue_put( res, trp_cons( trp_cons( trp_double( feat->x ), trp_double( feat->y ) ),
                                              trp_cons( trp_double( nbrs[0]->x ), trp_double( nbrs[0]->y ) ) ) );
                feat1[i].fwd_match = nbrs[0];
            }
        }
        free( nbrs );
    }
    kdtree_release( kd_root );
    return res;
}



