/*
    TreeP Run Time Support
    Copyright (C) 2008-2012 Frank Sinapsi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../trp/trp.h"
#include "./trpmagic.h"
#include <magic.h>

#define trp_magic_lock() (void)pthread_mutex_lock( &_trp_magic_mutex )
#define trp_magic_unlock() (void)pthread_mutex_unlock( &_trp_magic_mutex )
static pthread_mutex_t _trp_magic_mutex = PTHREAD_MUTEX_INITIALIZER;
static magic_t _trp_magic;

static trp_obj_t *trp_magic_internal_unlock( const uns8b *m );

uns8b trp_magic_init()
{
    if ( _trp_magic = magic_open( MAGIC_CONTINUE ) )
        if ( magic_load( _trp_magic, NULL ) )
            if ( magic_load( _trp_magic, "magic.mgc" ) ) {
                magic_close( _trp_magic );
                _trp_magic = NULL;
            }
    return 0;
}

void trp_magic_quit()
{
    trp_magic_lock();
    if ( _trp_magic ) {
        magic_close( _trp_magic );
        _trp_magic = NULL;
    }
    trp_magic_unlock();
}

trp_obj_t *trp_magic_available()
{
    trp_obj_t *res;

    trp_magic_lock();
    res = _trp_magic ? TRP_TRUE : TRP_FALSE;
    trp_magic_unlock();
    return res;
}

static trp_obj_t *trp_magic_internal_unlock( const uns8b *m )
{
    trp_obj_t *res;
    uns8b *r;
    uns32b l;

    if ( m == NULL ) {
        trp_magic_unlock();
        return UNDEF;
    }
    l = strlen( m );
    r = trp_gc_malloc( l + 1 );
    (void)memcpy( r, m, l );
    trp_magic_unlock();
    for ( ; ; ) {
        if ( l == 0 ) {
            res = trp_cord( "data" );
            break;
        }
        l--;
        if ( ( r[ l ] != ' ' ) &&
             ( r[ l ] != '\t' ) &&
             ( r[ l ] != '\r' ) &&
             ( r[ l ] != '\n' ) ) {
            r[ l + 1 ] = '\0';
            res = trp_cord( r );
            break;
        }
    }
    trp_gc_free( r );
    return res;
}

trp_obj_t *trp_magic_file( trp_obj_t *path )
{
    uns8b *s;
    trp_obj_t *res = UNDEF;
    uns32b sl;

    trp_magic_lock();
    if ( _trp_magic == NULL ) {
        trp_magic_unlock();
        return UNDEF;
    }
    s = trp_csprint( path );
    sl = strlen( s );
    if ( sl >= 4 ) {
        if ( ( strcmp( s + sl - 4, ".trp" ) == 0 ) ||
             ( strcmp( s + sl - 4, ".tin" ) == 0 ) )
            res = trp_cord( "TreeP source code text" );
        if ( strcmp( s + sl - 4, ".lyx" ) == 0 )
            res = trp_cord( "LyX source code text" );
    }
    if ( res == UNDEF ) {
        FILE *fp = trp_fopen( s, "rb" );

        if ( fp ) {
            res = trp_magic_internal_unlock( magic_descriptor( _trp_magic, fileno( fp ) ) );
            fclose( fp );
        } else
            trp_magic_unlock();
    } else
        trp_magic_unlock();
    trp_csprint_free( s );
    return res;
}

trp_obj_t *trp_magic_buffer( trp_obj_t *raw, trp_obj_t *cnt )
{
    uns32b c;

    if ( raw->tipo != TRP_RAW )
        return UNDEF;
    if ( cnt ) {
        if (  trp_cast_uns32b( cnt, &c ) )
            return UNDEF;
        if ( c > ((trp_raw_t *)raw)->len )
            c = ((trp_raw_t *)raw)->len;
    } else {
        c = ((trp_raw_t *)raw)->len;
    }
    if ( c == 0 )
        return UNDEF;
    trp_magic_lock();
    if ( _trp_magic == NULL ) {
        trp_magic_unlock();
        return UNDEF;
    }
    return trp_magic_internal_unlock( magic_buffer( _trp_magic,
                                                    ((trp_raw_t *)raw)->data,
                                                    c ) );
}



